from typing import TypeVar
from .base import BaseHttpService, ServiceError
from .deserializers import raw_deserializer, default_deserializer\

T = TypeVar("T")

class QueueService(BaseHttpService):
    def __init__(self, baseUrl: str) -> None:
        super().__init__(baseUrl)

    async def queues(self) -> list[str]:
        return await self.get("queues/")
    
    async def dequeue(self, queue_name) -> str|None:
        data:str = await self.get(f"queues/{queue_name}", caster=raw_deserializer)
        return data if len(data) > 0 else None
    
    async def enqueue(self, queue_name, data:T):
        return await self.post(f"queues/{queue_name}", data, is_raw=True)
    
    async def delete_queue(self, queue_name) -> bool:
        try:
            await self.delete(f"queues/{queue_name}")
        except ServiceError as e:
            return False


from typing import TypeVar, Callable
from enum import IntEnum
import requests
import asyncio

from .deserializers import default_deserializer

T = TypeVar("T")
TResult = TypeVar("TResult")

class ResultStatus(IntEnum):
    Unknown = 0
    Information = 1
    Success = 2
    Redirect = 3
    ClientError = 4
    ServerError = 5

def get_result_status(status_code:int) -> ResultStatus:
    match status_code:
        case x if x < 100:
            return ResultStatus.Unknown
        case x if x < 200:
            return ResultStatus.Information
        case x if x < 300:
            return ResultStatus.Success
        case x if x < 400:
            return ResultStatus.Redirect
        case x if x < 500:
            return ResultStatus.ClientError
        case _:
            return ResultStatus.ServerError

class ServiceError(Exception):
    def __init__(self, status:ResultStatus, *args: object) -> None:
        super().__init__(*args)
        self.status = status

class BaseHttpService:
    def __init__(self, baseUrl:str) -> None:
        self.baseUrl = baseUrl

    async def get(self, url:str, headers:dict[str, str] = None, caster:Callable[[str,], T] = None, stream:bool = False) -> T:
        caster = caster or default_deserializer
        request = lambda : requests.get(f"{self.baseUrl}/{url}", headers=headers, stream=stream)
        response = await asyncio.to_thread(request)
        status = get_result_status(response.status_code)
        if status == ResultStatus.Success:
            return caster(response.text)
        else:
            raise ServiceError(status, response.text)

    async def post(self, url:str, parameter:T, 
             headers:dict[str, str] = None, caster:Callable[[str,], TResult] = None,
             is_raw:bool = False
             ) -> TResult:
        caster = caster or default_deserializer
        params = {
            "headers": headers,
            "data" if is_raw else "json":parameter
        }
        request = lambda : requests.post(f"{self.baseUrl}/{url}",  **params)
        response = await asyncio.to_thread(request)
        status = get_result_status(response.status_code)
        if status == ResultStatus.Success:
            return caster(response.text)
        else:
            raise ServiceError(status, response.text)
        
    async def delete(self, url:str, headers:dict[str, str] = None, caster:Callable[[str,], T] = None) -> T:
        caster = caster or default_deserializer
        print(f"{self.baseUrl}/{url}")
        request = lambda : requests.delete(f"{self.baseUrl}/{url}", headers=headers)
        response = await asyncio.to_thread(request)
        status = get_result_status(response.status_code)
        if status == ResultStatus.Success:
            return caster(response.text)
        else:
            raise ServiceError(status, response.text)


from typing import TypeVar, Any
from .base import BaseHttpService, ServiceError

class TreeService(BaseHttpService):
    def __init__(self, baseUrl: str) -> None:
        super().__init__(baseUrl)

    async def get_forests(self) -> list[str]:
        return await self.get("trees/")

    async def get_trees(self, forest:str) -> list[str]:
        return await self.get(f"trees/{forest}")

    async def index(self, forest:str, tree:str, path:str = "$"):
        return await self.get(f"trees/{forest}/{tree}/index:{path}")

    async def get_value(self, forest:str, tree:str, path:str = "$"):
        return await self.get(f"trees/{forest}/{tree}/{path}")

    async def set_value(self, forest:str, tree:str, path:str, value):
        return await self.post(f"trees/{forest}/{tree}/{path}", value)

    async def remove_value(self, forest:str, tree:str, path:str):
        return await self.delete(f"trees/{forest}/{tree}/{path}")


from typing import TypeVar, Any
from .base import BaseHttpService, ServiceError

class TablesService(BaseHttpService):
    def __init__(self, baseUrl: str) -> None:
        super().__init__(baseUrl)

    async def get_databases(self) -> list[str]:
        return await self.get('tables/')
    
    async def get_tables(self, database:str) -> list[str]:
        return await self.get(f'tables/{database}')
    
    async def create_table(self, database:str, table:str, **fields) -> bool:
        fields = ', '.join([f'{field} {fields[field]}' for field in fields])
        query = f'create table {table} ({fields});'
        data = await self.post(f'tables/{database}', query, is_raw=True)
        if data is None:
            return False
        return True
    
    async def insert(self, database:str, table:str, **item):
        fields = ', '.join([i for i in item])
        values = ', '.join([(f"'{item[i]}'" if isinstance(item[i], str) else str(item[i])) for i in item])
        query = f'insert into {table} ({fields}) values ({values});'
        if await self.post(f'tables/{database}', query, is_raw=True) is None:
            return False
        return True
    
    async def insert_many(self, database:str, table:str, items:list[dict[str, Any]]):
        return [await self.insert(database, table, **item) for item in items]
    
    async def drop_table(self, database:str, table:str) -> bool:
        if await self.post(f'tables/{database}', f'drop table {table};', is_raw=True) is None:
            return False
        return True

    async def execute_query(self, database:str, query:str):
        return await self.post(f'tables/{database}', query, is_raw=True)
    
    async def get_rows(self, database:str, table:str, page_number:int = 0, page_size:int = 10):
        return await self.get(f'tables/{database}/{table}?page_number={page_number}&page_size={page_size}')


from typing import TypeVar, Any
from .base import BaseHttpService, ServiceError
from .deserializers import raw_deserializer
import base64

T = TypeVar("T")

class BlobsIndex:
    def __init__(self, status:int, path:list[str], files:list[str], folders:list[str]) -> None:
        self.status:int = status
        self.path:list[str] = path
        self.files:list[str] = files
        self.folders:list[str] = folders

class BlobsService(BaseHttpService):
    def __init__(self, baseUrl: str) -> None:
        super().__init__(baseUrl)

    async def index(self, path:str) -> BlobsIndex:
        index = (await self.get(f"blobs/{path}"))["index"]
        return BlobsIndex(**index)
    
    async def read(self, path:str) -> str|None:
        try:
            b64content:str = await self.get(f"blobs/raw:{path}", caster=raw_deserializer)
            return base64.b64decode(b64content).decode()
        except:
            return None
    
    async def write(self, path:str, content:str) -> str:
        b64content =  base64.b64encode(content.encode()).decode()
        return await self.post(f"blobs/{path}", b64content)

    async def delete_blob(self, path:str) -> bool:
        try:
            await self.delete(f"blobs/{path}")
            return True
        except ServiceError as e:
            return False


from .blobs import BlobsService
from .tables import TablesService
from .queues import QueueService
from .trees import TreeService

class BureaucratConnection:
    def __init__(self, url:str) -> None:
        self.url = url
        self.queues:QueueService = QueueService(url)
        self.blobs:BlobsService = BlobsService(url)
        self.tables:TablesService = TablesService(url)
        self.trees:TreeService = TreeService(url)